/*
 * Zaptel sysfs support for Zapata Telephony interface
 *
 * Written by Maxime MENEAULT <maxmeneault@free.fr>
 *
 * From an idea of Tzafrir Cohen (Xorcom Ltd.)
 *
 * Copyleft (>) 2007 Maxime MENEAULT <maxmeneault@free.fr>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* Everything is no-op if udev not selected */
#include "zaptel-sysfs.h"
#ifdef CONFIG_ZAP_UDEV
/*
* DESIGN CONSIDERATIONS:
*
* To be compliant with all the 2.6 kernels is a real headache for
* the Linux device model.
* - Either we break the Linux device model implementing it in sysfs directly.
*	The maindrawback would be the duplication of code and code maintainability
*   (if we wanted to keep it in sync with it).
*
* - Or we accept some limitations of older kernels (that's the right approach).
*
* At the beginning I wanted to make something like:
*	zaptel/ (class)
*		channels/  (zt_device)
*					1/ (symlink to spans/1/channels/1)
*					2/ (symlink to spans/2/channels/1)
*					3/ (symlink to spans/2/channels/2)
*		spans/	(zt_device)
*				1/	(zt_device)
*					channels/	(fake zt_device or simple sysfs dir)
*								1/	(zt_device)
*				2/	(zt_device)
*					channels/	(fake zt_device or simple sysfs dir)
*								1/ (zt_device)
*								2/ (zt_device)
*
* zt_device: it denotes either a class_device or device (depending on your kernel
*      version).
*
* Note that I have only output the main elements here (not the ones
* related to linux device model and neither the attributes of each kobject).
*
* This design is not particularly good however:
* - channels/ will break in some way the linux device model
* - It wouldn't be possible to implement under 2.6.15 kernels
* (without duplicating linux device model code -- class_device nesting
* was added in 2.6.15).
* - It will break existing versions of udev and every user wanting to use this new version
* of zaptel would have to rewrite its udev rules (well, at least install our 
* new udev rules)
* - Moreover there could be utilities relying on the old /sys/class/zap%n
* (don't want to break them neither).
*
*
* So I will provide the same functionnality but into a different design:
*
*	zaptel/ (class) 
*			zap1/ (zt_device)
*				span/ (symlink to span1)
*			zap2/ (zt_device)
*				span/ (symlink to span2)
*			zap3/ (zt_device)
*				span/ (symlink to span2)
*			span1/ (zt_device)
*				  chan1/ (symlink to zap1)		
*			span2/ (zt_device)
*				 chan1/	(symlink to zap2)
*				 chan2/ (symlink to zap3)
*
* NOTE: For kernels newer than 2.6.23 everything will be
* set as symlinks pointing under /sys/devices.
*
*
* LOCKING POLICY :
* At the time of this writting zaptel-base still needs fixes from the
* locking point of view but we made the following assumptions on the
* locking policy.
*
*	- every channel access must be protected with chan->lock held
*	- holding chan_lock does not prevent from getting chan->lock!
*	- locking order first:chan_lock second:chan->lock
*	- span->lock and chan->lock are best held separately.
*   
*/



/* === Handy macros and tricks === */

/*
* If you can't find out where some identifiers are defined then
* check this "private" header out.
* It handles all the gorry details about linux device model 
* so as to get a clean approach of the problem.
*/
#include "zaptel-sysfs-private.h"




/* === Declarations === */
static ZT_UEVENT_CALLBACK(zt_sysfs_class_uevent, device);

/* === Global Variables === */
                               
/* the main zaptel class */
static struct class zt_sysfs_class =
{
	.name = "zaptel",
	CLASS_OWNER(.owner = THIS_MODULE,)
	.class_release = zt_sysfs_class_release_nop,
	.CLASS_RELEASE_FIELD = zt_sysfs_device_release,
	.CLASS_UEVENT_FIELD = zt_sysfs_class_uevent,
};

/* some devices that have are known statically */
static struct zt_sysfs_device static_devices[] =
{
	ZT_DEFINE_STATIC_DEVICE(zaptranscode, MKDEV(ZT_MAJOR, 250)),
	ZT_DEFINE_STATIC_DEVICE(zaptimer, MKDEV(ZT_MAJOR, 253)),
	ZT_DEFINE_STATIC_DEVICE(zapchannel, MKDEV(ZT_MAJOR, 254)),
	ZT_DEFINE_STATIC_DEVICE(zappseudo, MKDEV(ZT_MAJOR, 255)),
	ZT_DEFINE_STATIC_DEVICE(zapctl, MKDEV(ZT_MAJOR, 0)),
	ZT_STATIC_DEVICE_NULL,
};


/* == common attributes for channels == */

/* Here you may declare your own show/store functions
* or simply use handy macros to get this defined for you
*
* @note Only put declarations here not definitions this way we keep this part
* of the code readable and clean.
*/
ZT_ATTR_CHAN_SHOW_MEMBER(name, char *, "%s"); 
ZT_ATTR_CHAN_SHOW_MEMBER(channo, int, "%d");
ZT_ATTR_CHAN_SHOW_MEMBER(chanpos, int, "%d");
ZT_ATTR_CHAN_SHOW_MEMBER(sigcap, int, "%04X");
ZT_ATTR_CHAN_SHOW_MEMBER(sig, int, "%04X");

/* here a more complex example (channel's current law) */
ZT_ATTR_DEFINE_SHOW_SIMPLE(xlaw, chan, CHAN, int, "%u", 
	((chan->xlaw == __zt_alaw)? ZT_LAW_ALAW : ZT_LAW_MULAW));

/* channel's master is still harder we'll have to provide our
* own show function
*/
static ATTR_SHOW_FUNC(ZT_ATTR_GET_SHOW(masterno, chan), device, buf);

/* Here you may add your attributes definitions 
* either using handy macros or defining them at hand.
*
* @note some handy macros of linux kernel like:
*  - __ATTR from sysfs.h did not exist till 2.6.8 kernels
*	(you should provide a replacement for older kernels if you want to use them)
*/
static struct zt_device_attribute chan_attrs[] =
{
	ZT_ATTR_CHAN_RO(name),
	ZT_ATTR_CHAN_RO(channo),
	ZT_ATTR_CHAN_RO(chanpos),
	ZT_ATTR_CHAN_RO(sigcap),
	ZT_ATTR_CHAN_RO(sig),
	ZT_ATTR_CHAN_RO(xlaw),
	ZT_ATTR_CHAN_RO(masterno),
	ZT_ATTR_NULL,
};

/* == common attributes for spans == */

/* @note read the notes about channels attributes (above)*/

ZT_ATTR_SPAN_SHOW_MEMBER(name, char *, "%s");
ZT_ATTR_SPAN_SHOW_MEMBER(desc, char *, "%s");
ZT_ATTR_SPAN_SHOW_MEMBER(spanno, int, "%d");
ZT_ATTR_SPAN_SHOW_MEMBER(channels, int, "%u");
ZT_ATTR_SPAN_SHOW_MEMBER(alarms, int, "%d");


static struct zt_device_attribute span_attrs[] =
{
	ZT_ATTR_SPAN_RO(name),
	ZT_ATTR_SPAN_RO(desc),
	ZT_ATTR_SPAN_RO(spanno),
	ZT_ATTR_SPAN_RO(channels),
	ZT_ATTR_SPAN_RO(alarms),
	ZT_ATTR_NULL,
};


/* === private functions === */
/*
* Here you may define your own show/store functions for class_device
* attributes.
*
* If this file becomes to big we could move channel specific information
* into zaptel-sysfs-chan.c and span specific information into
* zaptel-sysfs-span.c.
*
*/

/* @brief print the channel's master number */
static ATTR_SHOW_FUNC(ZT_ATTR_GET_SHOW(masterno, chan), device, buf)
{
	struct zt_chan *chan = (struct zt_chan *)device->DEVICE_DATA_FIELD;
	struct zt_chan *master;
	int channo;
	unsigned long flags;
    
	if (!chan) {
		return 0;
	}
    
	spin_lock_irqsave(&chan->lock, flags);
   
	if (ZT_INVALID_CHANNO(chan->channo)) {
		spin_unlock_irqrestore(&chan->lock, flags);
		return 0;
	}
    
	master = chan->master;
	channo = chan->channo;
	spin_unlock_irqrestore(&chan->lock, flags);
    
	if (master == chan) {
		return sprintf(buf, "%d", channo);
	} else if (!master) {
		return 0;
    }
    
	spin_lock_irqsave(&master->lock, flags);
    
	if (ZT_INVALID_CHANNO(master->channo)) {
		spin_unlock_irqrestore(&master->lock, flags);
		return 0;
	}
    
	channo = master->channo;
	spin_unlock_irqrestore(&master->lock, flags);     
        
	return sprintf(buf, "%d", channo);
}                                               


/* === Uevent handling === */

/*
* @brief add variables to uevents happening on zaptel class behalf 
* 
* Here you may add any variables to be added depending on the type
* of the class device (default/channel/span).
*
* @note uevents have got a limited space for variables so
* only add the necessary variables and no more.
*
* @note these uevents may be of any kind (add/remove...)
*
* @note Moreover traditionnal 'udev' let you match sysfs attributes
* so you may not have to deal with uevent environment variable but
* rather with sysfs attributes
*/
static ZT_UEVENT_CALLBACK(zt_sysfs_class_uevent, device)
{
	ZT_UEVENT_VAR_DECLS
	struct zt_sysfs_device *zdev;
	
	if (!device)
		return -EINVAL;

	zdev = to_zt_sysfs_device(device);

	/*
	* the only variable that might be usefull is its type
	* (default/channel/span)
	*/
	ZT_UEVENT_VAR_ADD("TYPE", "%u", zdev->type);
                
	ZT_UEVENT_VAR_END
}

/* === public functions for zaptel driver === */

/* for their description take a look at zaptel-sysfs.h 
* usually you wouldn't have to modify them unless there is a bug
* or you want to change the layout of the sysfs tree.
*/


/*
* a span doesn't have any meaning for now as a device node
* so we don't assign it a dev_t type
*/
int zt_sysfs_span_register(struct zt_span *span)
{
	struct zt_sysfs_device *zdev;
	unsigned long flags;
	int spanno;
	struct device *device;
	int i = 0;
	int retval;
    
	if (!span)
		return -EINVAL;
        
	spin_lock_irqsave(&span->lock, flags);
    
	spanno = span->spanno;
    
	if (ZT_INVALID_SPANNO(spanno)) {
		spin_unlock_irqrestore(&span->lock, flags);
		return -ENXIO;
	}
    
	device = span->device;
	spin_unlock_irqrestore(&span->lock, flags);
    
	zdev = zt_sysfs_device_create(device, MKDEV(0,0), span, 
		ZT_DEVICE_TYPE_SPAN, "span%d", spanno);
   
	retval = ZT_PTR_ERR(zdev);
    
	if (retval)
		return retval;	
  
    
	while (span_attrs[i].attr.name) {
		retval = zt_device_create_file(&zdev->device, &span_attrs[i]);
		if (retval)
			goto error;
		i++;
	}
        
	spin_lock_irqsave(&span->lock, flags);
	span->sysfs_dev = zdev;
	spin_unlock_irqrestore(&span->lock, flags);
	
	return 0;

error:
	i--;
	while (i >= 0) {
		zt_device_remove_file(&zdev->device, &span_attrs[i--]);
	}
	zt_device_unregister(zdev);
    
	return retval;
}



void zt_sysfs_span_unregister(struct zt_span *span)
{
	struct zt_sysfs_device *zdev = NULL;
	unsigned long flags;
	int i = 0;
    
	if (!span)
		return;
        
	spin_lock_irqsave(&span->lock, flags);
    
	if (span->sysfs_dev) {
		zdev = span->sysfs_dev;
		span->sysfs_dev = NULL;
	}
	spin_unlock_irqrestore(&span->lock, flags);
    
	if (zdev) {
    
		/* usually it is best to remove in reverse order that the creation order
		* but for that case it does not hurt
		*/
		while (span_attrs[i].attr.name) {
			zt_device_remove_file(&zdev->device, &span_attrs[i++]);
		}
    
		zt_device_unregister(zdev);
	}
}


int zt_sysfs_channel_register(struct zt_chan *chan)
{
	struct zt_sysfs_device *zdev, *zdevsp;
	unsigned long flags;
	dev_t devt;
	int channo;
	int chanpos;
	struct zt_span *span;
	struct device *device;
	int i = 0;
	int retval;
	char temp[25]; /* enough for 4 chars & 64-bit number & null-char */
    
	if (!chan)
		return -EINVAL;
        
	spin_lock_irqsave(&chan->lock, flags);
    
	channo = chan->channo;
    
	if (ZT_INVALID_CHANNO(channo)) {
		spin_unlock_irqrestore(&chan->lock, flags);
		return -ENXIO;
	}
    
	devt = MKDEV(ZT_MAJOR, channo);
	span = chan->span;
	chanpos = chan->chanpos;
	spin_unlock_irqrestore(&chan->lock, flags);
    
	/* device has to be fetched from span */
	if (!span)
		return -ENXIO;
        
	spin_lock_irqsave(&span->lock, flags);
        
	if (ZT_INVALID_SPANNO(span->spanno)) {
		spin_unlock_irqrestore(&span->lock, flags);
		return -ENXIO;
	}
        
	device = span->device;
	zdevsp = span->sysfs_dev;
	spin_unlock_irqrestore(&span->lock, flags);
        
        
	zdev = zt_sysfs_device_create(device, devt, chan,
		ZT_DEVICE_TYPE_CHANNEL, "zap%d", channo);

	retval = ZT_PTR_ERR(zdev);
    
	if (retval)
		return retval;
   	
	while (chan_attrs[i].attr.name) {
		retval = zt_device_create_file(&zdev->device, &chan_attrs[i]);
		if (retval)
			goto error;
		i++;
	}
    
	/* we can now register a symlink to this channel in the span directory */
	if (zdevsp) {
		sprintf(temp, "chan%d", chanpos); 
		retval = sysfs_create_link(&zdevsp->device.kobj, &zdev->device.kobj,
			temp);
			
		if (retval)
			goto error;
            
		/* and we can provide a symlink to the span in the channel directory 
		* (we provide "span" in temp rather as static memory because the prototype
		*  doesn't specify the keyword 'const')
		*/
		sprintf(temp, "span");
		retval = sysfs_create_link(&zdev->device.kobj, &zdevsp->device.kobj,
			temp);
        
		if (retval)
			goto error2;
        
	}
           
	spin_lock_irqsave(&chan->lock, flags);
	chan->sysfs_dev = zdev;
	spin_unlock_irqrestore(&chan->lock, flags);
	
	return 0;

error2:
	sprintf(temp, "chan%d", chanpos);
	sysfs_remove_link(&zdevsp->device.kobj, temp);
error:
	i--;
	while (i >= 0) {
		zt_device_remove_file(&zdev->device, &chan_attrs[i--]);
	}
	zt_device_unregister(zdev);
	return retval;
}


void zt_sysfs_channel_unregister(struct zt_chan *chan)
{
	struct zt_sysfs_device *zdev = NULL, *zdevsp = NULL;
	unsigned long flags;
	int channo;
	int chanpos;
	int i = 0;
	char temp[25];
	struct zt_span *span;
    
	if (!chan)
		return;
        
	spin_lock_irqsave(&chan->lock, flags);
    
	channo = chan->channo;
    
	/* we don't even care here if channel is still
	* valid or not, we try to be as robust as possible
	* so as to release everything we allocated
	*/
    
	chanpos = chan->chanpos;
    
	if (chan->sysfs_dev) {
		zdev = chan->sysfs_dev;
		chan->sysfs_dev = NULL;
	}
	span = chan->span;
	spin_unlock_irqrestore(&chan->lock, flags);
    
	sprintf(temp, "span");
	sysfs_remove_link(&zdev->device.kobj, temp);
    
	/* 
	* likewise try to be robust
	*/
	if (span) {
		spin_lock_irqsave(&span->lock, flags);   
		zdevsp = span->sysfs_dev;
		spin_unlock_irqrestore(&span->lock, flags);
	}
    
	if (zdevsp) {
		sprintf(temp, "chan%d", chanpos);
		sysfs_remove_link(&zdevsp->device.kobj, temp);
	}
        
	if (zdev) {
    	
		/* usually it is best to remove in reverse order that the creation order
		* but for that case it does not hurt
		*/
		while (chan_attrs[i].attr.name) {
			zt_device_remove_file(&zdev->device, &chan_attrs[i++]);
		}
    
		zt_device_unregister(zdev);
	}
    
}


int zt_sysfs_init(void)
{
	int retval;
	int i = 0;
    
	retval = class_register(&zt_sysfs_class);
    
	if (retval)
		return retval;
    
	while(static_devices[i].device.class) {
		retval = zt_device_register(&static_devices[i]);
		if (retval)
			goto error;
		i++;
	}
	return 0;
    
error:
	i--;
	while (i >= 0) {
		zt_device_unregister(&static_devices[i--]);
	}
	class_unregister(&zt_sysfs_class); 
	return retval;
}



void zt_sysfs_destroy(void)
{
	int i = 0;
    
	while(static_devices[i].device.class) {
		zt_device_unregister(&static_devices[i++]);
	}
	class_unregister(&zt_sysfs_class);
}


#endif /* CONFIG_ZAP_UDEV */
